package ElchemeaCGI;
#$Id: ElchemeaCGI.pm,v 1.31 2015/08/24 07:01:46 sofc Exp $

#################################################################################
#
# Package ElchemeaCGI
#
# This module contains a number of utility functions for outputting properly formatted html code for
# user interface generation. Thus it mainly extends the CGI.pm module by Lincoln D. Stein.
# The module exports these functions in two groups.
#
#
# The :html group exports these functions:
#
# -- \begin{longtable}{ l p{10cm} }
# -- print_header($[\%]) & Prints the header information. Arguments: title. Any additional optional arguments (in the form of a hash) will be parsed along to the header() function supplied by CGI.pm. The functionautomatically appends a call to a javascript function logging users out after some time of no actions. \\
# -- print_end() & Prints the help button and ends the html output with the proper tag.  \\
# -- not_auth() & Prints the information supplied to the user if the user is nor authorised. Also prints a link to the log in page. \\
# -- print_hidden() & Prints a number of hidden fields used to maintain state. This includes user name and a cryptographic hash of the users password. \\
# -- print_hidden_rig() &  Same as print_hidden(), but with the additional information abut the active rig.\\
# -- logout() & Prints a logout button. \\
# -- action($) & Prints a hidden field with an action parameter with the specified value which can be used for program control flow.  \\
# -- EFA_start_html() & A wrapper for CGI::start_html. Any arguments (in the form of a hash) are passed to CGI::start_html. Automatically appends a reference to the javascript source file on the server. \\
# -- js_back() & Prints the javascript for gping backwards (uses the browser.back() fjavascrpt call). \\
# -- get_CGI_value($) & Retrieves the value of the specified CGI parameter (supplied by the web browser. \\
# -- get_CGI_value_clean($) & Same as get_CGI_value, but does pattern match on the retrieved value and only returns the part that matches. The pattern match is [{\backslash}w{\backslash}s{\backslash}.{\backslash},]*. This has the benefit of untainting the returned parameter value (For taint checks in perl and web access, refer Lincoln D. Steins book \textit{Official Guide to Programming with CGI.pm}) \\
# --\end{longtable}
#
#
# The :cgi group of functions include the following:
#
# --\begin{longtable}{ l p{10cm} }
# -- get_CGI_value($) & See above. \\
# -- get_CGI_value_clean($) &  See above. \\
# -- action($) & See above \\
# -- not_auth() & See above \\
# -- login_ok() & Checks if the user supplied login credentials are ok. This can be either against a small local database, or against a full RDBMS. \\
# -- login_msg() & returns the errormessage if the user was not authenticated.\\
# -- menu_button(@) & Prints a menu button. Arguments: name, value, style. The name will be the CGI parameter name, the value vill be the text on the button and the style is a style class name to use for displaying. \\
# -- create_menu_field & Prints the html tags to create a menu field.  \\
# -- top_nav_bar_start() & Prints the html tags to start the top navigation bar (table specifications etc.) \\
# -- top_no_button() & Prints a no action button (goes nowhere) in the top navigation bar. \\
# -- top_nav_bar_button() & prints a top navigation button. Arguments: File, name, value, style, [optional additional name, value and force triplets]. The file is the cgi-script to be called upon button press, the name,value and style arguments are passed to menu_button() and the additional optional arguments are used to initialise and print hidden html fields in the form of name-value pairs and a force argument (1 for force value, 0 for allow reuse of value). \\
# -- tab_newrow() & Prints a new row in the top navigation bar. \\
# -- top_js_return() & Prints a top navigation return button (uses the javascript printed by js_back(), see above) \\
# -- end_top_bar() & Prints the end of the top navigation bar. \\
# --\end{longtable}
#
#
#################################################################################



use strict;
use warnings;
use ElchemeaHeader;
use Elchemea;
use CGI qw(:standard :netscape);
require Exporter;

BEGIN {
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    @ISA = qw(Exporter);
    @EXPORT = qw($homedir $userdir);
    @EXPORT_OK = qw(top_nav_bar_button print_header print_end not_auth print_hidden logout login_ok action get_user EFA_start_html print_id js_back get_CGI_value get_CGI_value_clean js_back login_ok menu_button not_auth create_menu_field top_no_button tab_newrow action end_top_bar top_nav_bar_start top_js_return create_menu_field login_msg);
    %EXPORT_TAGS = (
		    html => [ qw(print_header print_end not_auth print_hidden logout action EFA_start_html print_id js_back get_CGI_value get_CGI_value_clean js_back) ],
		    cgi => [ qw(login_ok login_msg  get_CGI_value get_CGI_value_clean menu_button create_menu_field top_no_button tab_newrow action end_top_bar top_nav_bar_start top_js_return top_nav_bar_button not_auth) ],
                    );
}

my $msg = '';

return 1;

sub login_msg()
{
    return $msg;
}


sub login_ok()
{
    my $user = param('user');
    my $pass = param('pass');
    my $time = param('time');
    if (@_)
    {
	$user = shift;
	$pass = shift;
	$time = join('',Elchemea::get_time('+1min'));
    }
    return 0 unless ($user && $user =~ /\w+/);
    unless (Elchemea::user_exist($user))
    {
	$msg = "user does not exist on this Elchemea sustem!";
	return 0;
    }
    if (get_config_value('global','use_passwd') && get_config_value('global','use_passwd') =~ /yes/i)
    {
	if (get_config_value('global','passwd_server') && get_config_value('global','passwd_server') =~ /\w+/)
	{
	    my $host = get_config_value('global','passwd_server');
	    my $port = get_config_value('global','passwd_server_port');
	    my $cmd = get_config_value('global','passwd_server_auth_cmd');
	    unless (Elchemea::socket_client($host,$port,$cmd,$user,$pass))
	    {
		$msg = 'Wrong username and password combination reported by '.$host;
		return 0;		
	    }
	}
	else
	{
	    my $pwd = Elchemea::get_pwd($user);
	    unless ($pass eq $pwd)
	    {
		$msg = 'Wrong username and password combination';
		return 0;
	    }
	}
    }
    if (get_config_value('global','use_time') && get_config_value('global','use_time') =~ /yes/i)
    {
	my $now = join('',Elchemea::get_time('now'));
	unless ($time > $now)
	{
	    $msg = "Time too logn since last command, user logged out!";
	    return 0;
	}
    }
    $msg = '';
    return 1;
}

sub action($)
#############################################################
### This function prints a hidden field with hame action ####
### and value equal to the argument                      ####
### Usage  action($value);                               ####
#############################################################
{
    print hidden(-name=>'action',-value=>shift,-force=>'1');
}

sub login
################################################
### This function prints a login table      ####
### The html tags created is:               ####
### name: the username                      ####
### pass: The password                      ####
### action: The name of the submit button   ####
### The value of this is 'Login'            ####
################################################
{
    print "\n",'<script lanuage="javascript">',"\n";
    my $script = 'popup.html';
    if (get_config_value('global','use_passwd') && get_config_value('global','use_passwd') =~ /yes/)
    {
	$script = 'popup_pwd.html';
    }
    print <<END;

    function windowOpen() {
        var myWindow=window.open('../../elchemea/$script','windowRef','width=450,height=100,left=300,top=200');
        myWindow.location.href = '../../elchemea/$script';
    if (!myWindow.opener)
        myWindow.opener = self;
        myWindow.focus();
    }

    function go()
    {
        document.login.action.click();
    }

    windowOpen();
//--></script>

END

    print start_form(-name=>'login',-class=>"form",-onsubmit=>'return checkjs()',-action=>'login.cgi');
    print hidden(-name=>'name',-value=>'',-value=>'1');
    print hidden(-name=>'pass',-value=>'',-value=>'1');
    print submit(-name=>'action',-value=>'Login',-class=>'hidden');
    print '</th></tr></table>';
    print end_form;
}

sub EFA_start_html
#####################################################################
### This function is a wrapper for start_html but automatically #####
### if the onload argument is already given the logout script   #####
### is just added to the functions which are to be run.         #####
#####################################################################
{
    my %h = @_;
    my $onload = 0;
    for my $key (keys %h)
    {
        if ($key =~ /-onload/i)
        {
            my $str = $h{$key};
            my $e = '';
            if ($str =~ /([\'\"])$/)
            {
                $e = $1;
                $str =~ s/[\'\"]$//;
            }
            $str = $str . ';'.$e;
            $h{$key} = $str;
            $onload = 1;
        }
    }
    $h{'-script'} = { -language => 'javascript', -src => '/elchemea/Elchemea.js' };
    return CGI::start_html(%h);
}

sub print_header($)
#########################################################
### This function prints the header for the web pages ###
### Any second argument is used as the javascript     ###
### function to be called after the page loads        ###
#########################################################
{
    my $tit = shift;
    my $use_time = 0;
    $use_time = 1 if (get_config_value('global','use_time') && get_config_value('global','use_time') =~ /yes/i);
    my $res = header(-charset=>'utf-8');
    my %h = @_;
    my $onload = 0;
    for my $key (keys %h)
    {
        if ($key =~ /-onload/i)
        {
            my $str = $h{$key};
            my $e = '';
            if ($str =~ /([\'\"])$/)
            {
                $e = $1;
                $str =~ s/[\'\"]$//;
            }
            $str = $str . ';doAlertLogout();'.$e if ($use_time);
            $h{$key} = $str;
            $onload = 1;
        }
    }
    if ($use_time && !$onload)
    {
        $res .= EFA_start_html(%h,-title=>$tit,-style=>{src=>'/elchemea/standard.css',src=>'/elchemea/Elchemea.css'},-onLoad=>'doAlertLogout()');
    }
    else
    {
        $res .= EFA_start_html(%h,-title=>$tit,-style=>{src=>'/elchemea/standard.css',src=>'/elchemea/Elchemea.css'});
    }
    print $res;
    print "\n\n";
    print start_form(-name=>'login_info');
    print_hidden();
    print end_form();
    print "\n\n";
}

sub print_end()
#######################################
### Prints the end_html command #######
#######################################
{
    my $file = 'manual.pdf';
    $file = shift if (@_);
    print "\n\n";
    print p();
    print center(a({-href=>'javascript:load_help();'},'Help'));
    print <<END;

<script lanuage="javascript">
function load_help()
{
    var myWindow=window.open('/elchemea/$file','windowRef','width=700,height=500,left=100,top=100,resizable=yes,scrollbars=yes');
    myWindow.location.href = '/elchemea/$file';
    if (!myWindow.opener)
    {
        myWindow.opener = self;
        myWindow.focus();
    }
}

</script>

END
    print center(p('Copyright &copy '.$ElchemeaHeader::COPYYEAR.' '.$ElchemeaHeader::DESIGNERS.', '.$ElchemeaHeader::COPYRIGHT));
    print end_html;
}

sub logout
##########################################
###  Function printing a logout button ###
##########################################
{
    my $user = param('user');
    print start_form(-action=>'prelogin.cgi');
    if ($user)
    {
	print submit(-class=>'s0',-name=>'logout',-value=>'Logout '.$user);
    }
    else
    {
	print submit(-class=>'s0',-name=>'logout',-value=>'Logout');
    }
    print end_form;
}


sub not_auth()
{
    print_header("Access to Elchemea denied!");
    print h1({-class=>'C'},"Access to Elchemea denied!");
    if ($msg)
    {
	print h2({-class=>'C'},$msg);
    }
    else
    {
	print h2({-class=>'C'},"Possibly wrong username");
    }
    print '<center>';
    &create_menu_field;
    print start_form(-class=>"form",-name=>'Return',-action=>'prelogin.cgi');
    print '<table>';
    print '<tr><td width=200>';
    &menu_button("To login page","To login page",'s4');
    print '</td></tr></table>';
    print end_form();
    print '</center>';
    print_end;
}


sub js_back
##############################################
### This function prints a javascript which ##
### takes the browser one step backwards    ##
##############################################
{
    print "\n",'<script>',"\n";
    print 'function back() {',"\n";
    print '  history.back();',"\n";
    print '  return false;',"\n";
    print '}',"\n";
    print '</script>',"\n";
}

sub print_hidden()
#########################################################################
### Prints the logininformation to be transfered to the next cgi call ###
### If called with arguments the username and password is used from   ###
### the passed arguemnts instead from CGI parameters                  ###
#########################################################################
{
    my $user = param('user');
    my $pass = param('pass');
    if (@_)
    {
	$user = shift;
	$pass = shift;
    }
    my $time = join('',Elchemea::get_time(get_config_value('global','logoutdelay')));
    print hidden(-name=>'user',-value=>$user,-force=>'1');
    print hidden(-name=>'pass',-value=>$pass,-force=>'1');
    print hidden(-name=>'time',-value=>$time,-force=>'1');
}


sub top_nav_bar_start
##################################################
###Function creating the top_navigator_bar   #####
###                                          #####
##################################################
{
    print '<table /*border="2px"*/ class="top_table", width=100% >';
    print '<col width=20%", span="5">';
    print '<tr>';
    &top_logout();
    print '</td>';
    &print_page();
}

sub print_page()
###################################################
### Function creating a top bar button for      ###
### printing the active page                    ###
###################################################
{
    print '<td>';
    print "\n<script lanuage=\"javascript\">\n";
    print "function printWindow() {\n";
    print "   var bV = parseInt(navigator.appVersion);\n";
    print "   if (bV >= 4) {\n";
    print "      window.print();\n";
    print "   }\n";
    print "}\n";
    print "</script>\n";
    print start_form(-name=>'dummy',-class=>'form');
    print button(-name=>'Printpage',-value=>'Print page',-class=>'print',-onclick=>'printWindow()');
    print end_form;
    print '</td>';
}


sub top_nav_bar_button($$@)
#######################################################################
### Prints a nav bar button which when pressed                     ####
### takes the user to the $file CGI-script and                     ####
### parses the action parameter $action to the                     ####
### script. For the 3 last arguments, see                          ####
### menu_button()                                                  ####
### Usage top_nav_bar_button($file,$action,$1,$2,$3,[opt args])    ####
### Arguments 1 to 3 are parced directly to                        ####
### menu_button().                                                 ####
### It is possible to specify parameters to be printed in hidden   ####
### fields by adding additional arguments, each additional         ####
### argument must be a comma separated list specifying (in order!) ####
### name, value and force (0/1). if the parameter is just to be    ####
### 'sticky, let the last two elements be blank, eks: search_val,, ####
### would print a hidden field with name 'search_val' and the      ####
### value would be whatever the previous field with name           ####
### 'searh_val' had.                                               ####
#######################################################################
{
    my $file = shift;
    my $action = shift;
    &create_menu_field;
    if ($file)
    {
        print start_form(-action=>$file,-class=>"form");
    }
    else
    {
        print start_form(-class=>"form");
    }
    &action($action) if ($action);
    my @args = ();
    push @args, shift @_;
    push @args, shift @_;
    push @args, shift @_;
    &menu_button(@args);
    &print_hidden();
    while (@_)
    {
        my @s = split(/\,/,shift @_);
        print hidden(-name=>$s[0],-value=>$s[1],-force=>$s[2]);
    }
    print end_form();
    print '</td>';
}



sub top_no_button
#################################
### Prints a 'empty' button   ###
### in the top nav bar        ###
#################################
{
    &create_menu_field;
    start_form;
    &menu_button('','','none');
    end_form;
    print '</td>';
}

sub top_js_return
#############################################################
### Prints a return button  in the top navbar which uses  ###
### javascript to bring the user one step back            ###
### Note that the function requires that the function     ###
### &js_back() is used somewhere on the page!!!           ###
#############################################################
{
    &create_menu_field;
    print start_form(-class=>"form",-onsubmit=>"return back()");
    &print_hidden();
    &menu_button('Back','Back','back');
    print end_form();
    print '</td>';
}
sub top_logout
#############################################################
### Prints a logout button  in the top navbar which uses  ###
### javascript to ask if the user wants to logout         ###
#############################################################
{
    print "\n\n<script lanuage=\"javascript\">\n";
    print "IsLoggedIn = 1;\n\n";
    print "function LogOut() {\n";
    print "   if (IsLoggedIn == 0) {\n";
    print "       return true;\n";
    print "   }\n";
    print "   if (confirm('Do you want to log out?')) {\n";
    print "       return true;\n";
    print "   }\n";
    print "   return false;\n";
    print "}\n";
    print "</script>\n\n";
    my $user = param('user');
    &create_menu_field;
    print start_form(-class=>"form",-name=>'logout',-action=>'prelogin.cgi',-onsubmit=>"return LogOut()");
    if ($user)
    {
	&menu_button("Logout","Logout $user",'logout');
    }
    else
    {
	&menu_button("Logout","Logout",'logout');
    }
    print end_form();
    print '</td>';
}


sub tab_newrow
################################
### prints a html collumn    ###
### separator                ###
################################
{
    print '</tr><tr>';
}

sub end_top_bar
#################################
### Ends the top nav bar     ####
#################################
{
    print '</tr>';
    print '</table>';
}


sub create_menu_field
###################################################
### This function prints a table field with a   ###
### specific height                             ###
###################################################
{
    print '<td height="30px", margin="0px", padding="0px">';
}

sub menu_button($$$)
##################################################################
### This funtion creates a menu button                         ###
### Usage: menu_button($bgcolor,$color,$value,$name);          ###
##################################################################
{
    my $name = shift;
    my $value = shift;
    my $style = shift;
    print submit(-name=>$name,-class=>$style,-value=>$value);
}

sub clean($)
#######################################################
#### This function cleans a string for all special  ###
#### characters (only allowed ones are . and ,)     ###
#######################################################
{
    my $val = shift;
    my $ret = undef;
    if ($val =~ /([\w\s\.\,]*)/)
    {
        $ret = $1;
    }
    return $ret;
}

sub get_CGI_value_clean($)
##################################################
#### Wrapper for &clean and &get_CGI_value    ####
##################################################
{
    my $id = shift;
    my $val = &get_CGI_value($id);
    return &clean($val);
}

sub get_CGI_value($)
####################################################
#### This function gets the CGI parameter with   ###
#### $name, returns undef if the parameter is    ###
#### not defined in the input parameters         ###
#### Usage: get_CGI_value($name)                 ###
####################################################
{
    my $id = $_[0];
    my $val = undef;
    $val = param($id) if defined(param($id));
    return $val;
}
