package Solartron1287;

############################################################
#
#    Package Solartron1287
#
# This module implements the PSTAT class for Solartron
# --{\textregistered} 
# 1286 and 1287 electrochemical interfaces.
#
# To get a device instance of one of the classes call the constructors:
#
# $pstat = Solartron1286->new($gpib_address);
#
# $pstat = Solartron1287->new($gpib_address);
#
############################################################


use strict;
use warnings;
use FDEV;
our @ISA = qw(PSTAT);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1287';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_minpV'} = -14.5;
    $self->{'_maxpV'} = 14.5;
    $self->{'_minpA'} = -2;
    $self->{'_maxpA'} = 2;
    $self->{'_mins'} = 0.0001;
    $self->{'_maxs'} = 100;
    $self->{'_AC_gain'} = undef;
    $self->{'_setuptime'} = 4;
    $self->{'_overhead'} = 1.0;
    $self->{'_ranges'} = [0,0.1,2,20,50];
    $self->{'_resistors'} = [0,0.1,1,10,100,1000,10000,100000,1000000];
    $self->{'_freq_seg_tags'} = ['measure_resistor'];
    $self->{'_delay'} = 0.1;  ## Default delay of 100 ms before sweep ##
    $self->{'_setup_tags'} = 'mode,polarisation,polarisation_offset,current_limit,use_0.01_factor,bandwidth,range,bias_reject,measure_resistor,reference_potential,delay';
    $self->{'_setup_tags_chrono'} = 'mode,polarisation,polarisation_offset,current_limit,delay,bandwidth,reference_potential,measure_resistor,range';
    $self->{'_setup_tags_potsweep'} = 'mode,polarisation,current_limit,polarisation_offset,measure_resistor,sweep_rate,bandwidth,delay,reference_potential,range,sweeps,keep_potential';
    $self->{'_default'} = {
        'mode' => 'Volt',
	'polarisation_offset' => 'OFF',
	'keep_potential' => 'NO',
	'reference_potential' => 'Fixed',
	'measure_resistor' => 0.1,
	'use_0.01_factor' => 'YES',
	'bias_reject' => 'NO',
	'bandwidth' => 'C',
	'current_limit' => 2,
	'range' => 'auto',
	'sweeps' => 1,
	'delay' => $self->{'_delay'},
	'polarisation' => 0,
    };
    $self->{'_desc'} = {
        'mode' => 'Control mode, either voltage or current control',
	'polarisation_offset_program' => 'Determines if an external program may supply an offset to the specified polarisation, and if so which program',
	'keep_potential' => 'Determines if the potentiostat is left on (that is active) as opposed to standby (open circuit) after a chrono measurement has completed',
	'reference_potential' => 'Determines if the polarisation is set with respect to the potentiostat ground (fixed) or with respect to the sample rest potential (OCV). If OCV is selected, note that the potentiostat needs to go to OCV to determine the correct potential before it can be applied!',
	'measure_resistor' => 'Measure resistor for current measurements. Notice that the masximum potential allowed across the measure resistor is 200 mV!',
	'use_0.01_factor' => 'AC input gain of the potentiostat. If small amplitudes are to be used it is often advantegeous to use the 0.01 gain so larger AC output of the FRA can be used.',
	'bias_reject' => 'Allows the potentiostat to subtract the DC bias from the output leads to the FRA, potentially leading to better impedance results for polarised samples',
	'bandwidth' => 'Input filter type (refer the potentiostat manual for correct filter selection)',
	'current_limit' => 'Current limit, if current is cetected above this value the potentiostat switches to rest (OCV) conditions',
	'range' => 'Input range for voltage measurements',
	'sweeps' => 'Number of sweeps (cycles) in a singel potential sweep scan',
	'delay' => 'Delay in seconds before measuremnts starts',
	'sleeptime' => 'Delay in seconds between each data point',
	'sweep_rate' => 'Sweep rate in V/s or A/sec',
	'polarisation' => 'DC polarisation in either volts or amps',
	'bias_point_1' => 'Start polarisation',
	'bias_point_2' => 'First apex polarisation',
	'bias_point_3' => 'Second paex polarisation',
	'bias_point_4' => 'End polarisation',
	'totaltime' => 'Total chrono measure time',
    };
    $self->{'_values_measure_resistor'} = ['auto',0.1,1,10,100,1000,10000,100000,1000000];
    $self->{'_values_use_0.01_factor'} = ['YES','NO'];
    $self->{'_values_bias_reject'} = ['YES','NO'];
    $self->{'_values_keep_potential'} = ['YES','NO'];
    $self->{'_values_bandwidth'} = ['A','B','C','D','E','F','G','H','I','J'];
    $self->{'_values_current_limit'} = [0.000002,0.00002,0.0002,0.002,0.02,0.2,2];
    $self->{'_values_reference_potential'} = ['Fixed','OCV'];
    $self->{'_values_range'} = ['auto',0.1,1,2,20,50];
    $self->{'_save_tags'} = ['RG','FI','BY','ON','SY','PB','GB','RR','PO','PV','PC','DL','OL','GP','TR','BY','SW','IL','PX','PY','UL','UR','FD','PI','VT','IT','BR','PW','OF'];
    $self->{'_errorcodes'} = {
        '00' => ' ',
        '01' => 'Unknown command',
        '02' => 'Argument mismatch',
        '03' => 'Argument out of range',
        '04' => 'Floating point format error',
        '05' => 'Ilegal request for value',
        '20' => ' ',
        '21' => 'IRC not allowed in current (G-stat) mode',
        '22' => 'Pol I or I-reject not compatible with measure resistor',
        '23' => 'Missing sweep parameter',
        '24' => 'Zero or negative value in log/time plot',
        '25' => 'VX min > vX max and/or vY min > VY max in the plot',
        '26' => 'IRC or RPC inconsistent with selected resistor',
        '27' => 'GPIB-echo not allowed in plot',
        '28' => 'Ramp sweel: ramprate > 100 V/s',
        '29' => 'Ramp sweep: ramprate < 100uV/s',
        '30' => ' ',
        '31' => 'Current DVM overload',
        '32' => 'Voltage DVM overload',
        '33' => 'Both DVM overload',
        '34' => 'Thermal cut-out',
        '35' => 'RE1 overload',
        '36' => 'RE1-RE2 overload',
        '37' => 'Current overload',
        '38' => 'Current limit',
        '39' => 'ICE standby',
        'XX' => 'No data recieved',
    };
    $self->{'_max_sweep_points'} = 4;
    bless $self,$class;
    return $self;
};

sub bulk_load_check(@)
{
    my $self = shift;
    $self->writecmd('CE');
    $self->writecmd('OT1'); ## Terminator set to 'cr lf + EOI' ##
    while (@_)
    {
        my $cmd = shift;
        $self->checkcmd($cmd);
    }
    return undef;
}

sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
        $self->writecmd('?'.$tag);
        my $r = $self->readcmd();
        if ($r =~ /([\d\.\+\-eE\,]+)/)
        {
            push @res,$tag.' '.$1;
        }
    }
    return @res;
}

sub AC_gain()
{
    my $self = shift;
    return $self->{'_AC_gain'} if defined($self->{'_AC_gain'});
    $self->writecmd('?PI');
    my $res = $self->readcmd();
    if (defined($res) && $res =~ /01/)
    {
	$self->{'_AC_gain'} = 0.01;
    }
    elsif (defined($res) && $res =~ /00/)
    {
	$self->{'_AC_gain'} = 1;
    }
    return $self->{'_AC_gain'};
}

sub init()
{
    my $self = shift;
    $self->{'_pol_offset'} = 0;
    $self->writecmd('BK4'); ## Init ##
    sleep(5);
    $self->writecmd('OT3');
    $self->writecmd('GP2');
    $self->writecmd('OT1'); ## Terminator set to 'cr lf + EOI' ##
    $self->checkcmd('EB1'); ## Beep off ##
    $self->checkcmd('EB1'); ## Beep off, I mean it! ##
    $self->checkcmd('SW0'); ## sweep off ##
    $self->checkcmd('BY0'); ## Standby set to All o/c ##
    $self->checkcmd('ON1'); ## On mode set to Rest V/I ##
    $self->checkcmd('PW0'); ## Off ##
    $self->checkcmd('PI0'); ## gain x1 ###
    $self->writecmd('?PI');
    my $res = $self->readcmd();
    if (defined($res) && $res =~ /01/)
    {
	$self->{'_AC_gain'} = 0.01;
    }
    elsif (defined($res) && $res =~ /00/)
    {
	$self->{'_AC_gain'} = 1;
    }
    $self->writecmd('?VN');
    return $self->readcmd();
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('OT3');
    $self->writecmd('GP2');
    $self->writecmd('OT1'); ## Terminator set to 'cr lf + EOI' ##
    $self->writecmd('?TS');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /\d/);
    return 0;
}

sub reset()
{
    my $self = shift;
    $self->writecmd('BY1'); ## Standby set to CE o/c##
    $self->checkcmd('ON1'); ## On mode set to Rest V/I ##
    $self->writecmd('SW0'); ## sweep off ##
    $self->writecmd('EB1'); ## Beep off ##
    $self->writecmd('OT3');
    $self->writecmd('GP2');
    $self->writecmd('OT1');  ### Terminator set to 'cr lf + EOI' ###
    $self->checkcmd('PW0'); ## Off ##
}


sub set_polarisation($)
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
    if (ref($ext_pol) eq 'RemoteExec' && defined($pol) && $pol =~ /\d/i)
    {
	$self->writedebug("Executing external function to get polarisation offset");
	$self->{'_pol_offset'} = $ext_pol->exec_id($self);
	$pol += $self->{'_pol_offset'};
    }
    my $ocv = 0;
    $ocv = 1 if ($self->get_setup_value('Fixed',$setup,'reference_potential') =~ /OCV/i);
    $self->writecmd('OT3');
    $self->writecmd('GP2');
    $self->writecmd('OT1');  ### Terminator set to 'cr lf + EOI' ###
    if (!defined($pol) || $pol =~ /OCV/i)
    {
        ## On mode set to Rest V/I ##
	$self->checkcmd('ON 1');
    }
    else
    {
        ## On mode set to Pol V/I ##
	$self->checkcmd('ON 0');  
    }
    if ($self->{'_mode'} =~ /A/)
    {
	$self->writedebug('WARNING: Using OCV as reference potential is not possible in galvanostatic mode!') if ($ocv);
        $pol = $self->{'_minpA'} if ($pol < $self->{'_minpA'});
        $pol = $self->{'_maxpA'} if ($pol > $self->{'_maxpA'});
	$self->checkcmd('PO 1');
        $self->checkcmd(sprintf("PC %.4g",$pol));
    }
    else
    {
	if ($ocv)
	{
	    $self->writedebug('Determining OCV potential.');
	    $self->writecmd('BY 1'); ## Standby set to CE o/c##
	    $self->checkcmd('PW 0'); ## Off ##
	    sleep(1);
	    my ($v,$i) = split(/\s+/,$self->run_dvm());
	    $self->writedebug("OCV potential: $v V.");
	    $self->{'_last_ocv'} = $v;
	    $pol += $v unless (!defined($pol) || $pol =~ /OCV/i);
	}
	$pol = 0 if (!defined($pol) || $pol =~ /OCV/i);
	$pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
	$pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'});
	$self->checkcmd('PO 0');
        $self->checkcmd(sprintf("PV %.4g",$pol));
    }
    return $pol;
}

sub tag_value
{
    my $self = shift;
    my $key = shift;
    if ($key eq 'measure_resistor')
    {
	return $self->{'_resistor'} unless (@_);
	return $self->set_resistor(shift);
    }
}

sub set_resistor()
{
    my $self = shift;
    my $rr = shift;
    $rr = 0 if ($rr =~ /auto/i);
    if ($rr)
    {
	my @list = @{$self->{'_resistors'}};
	my $tr = 0;  ## Default auto resistor ##
	for (my $x = 0; $x < @list; $x++)
	{
	    next unless ($list[$x] == $rr);
	    $tr = $x;
	    $self->{'_resistor'} = $rr;
	}
	$rr = $tr;
    }
    ### Get existing resistor value ###
    $self->writecmd('?RR');
    my $res = $self->readcmd();
    my $oldr = undef;
    if ($res && $res =~ /(\d+)/)
    {
	$oldr = $1;
    }
    ### Only set new ressitor if nescesarry ###
    if (!defined($oldr) || $oldr != $rr)
    {
	$self->checkcmd("RR$rr");  ## Measure resistor ###
	sleep(1);  ## Allow instrument to setle on new ressitor value ## 
    }
    return $self->{'_resistor'};
}

sub sweep_time()
{
    my $self = shift;
    my $setup = shift;
    my $potlist = shift;
    die "Potential list not an arrayref\n" unless ($potlist && ref($potlist) =~ /ARRAY/);
    my @plist = @{$potlist};
    die "Too many independent potential points atttempted for potential sweep!\n" if (@plist > $self->{'_max_sweep_points'});
    my $sr = $self->get_setup_value(0,$setup,'sweep_rate');
    foreach my $pol (@plist)
    {
	if ($self->{'_mode'} =~ /A/)
	{
	    $pol = $self->{'_minpA'} if ($pol < $self->{'_minpA'});
	    $pol = $self->{'_maxpA'} if ($pol > $self->{'_maxpA'});
	}
	else
	{
	    $pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
	    $pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'}); 
	}
    }
    $sr = $self->{'_mins'} if ($sr < $self->{'_mins'});
    $sr = $self->{'_maxs'} if ($sr > $self->{'_maxs'});
    my $sp = $plist[0];
    my $totaltime = 0;
    my @tlist = ();
    for (my $x = 0; $x < scalar(@plist)-1; $x++)
    {
	my $t = abs($plist[$x+1] - $plist[$x]) / $sr;
	$t = 0.01 if ($t < 0.01);
	push @tlist,$t;
	$totaltime += $t;
    }
    if (@plist > 1)
    {
	push @tlist,0.01;
	$totaltime += 0.01;
    }
    my $sweeps = int($self->get_setup_value($self->get_default('sweeps'),$setup,'sweeps'));
    $totaltime = $self->{'_delay'} + $totaltime * $self->{'_overhead'} * $sweeps;
    return $totaltime;
}

sub square_time()
{
    my $self = shift;
    my $setup = shift;
    my $sweeps = 1500; ### output for 1 minute ###
    my $totaltime = $sweeps * 4 * 0.01 * $self->{'_overhead'}; 
    return $totaltime;
}

sub setup_square()
{
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    my $truepol = $self->setup_common($setup);
    my $ocv = $self->get_last_ocv();
    $ocv = 0 unless ($ocv);
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
#    if (ref($ext_pol) eq 'CODE')
#    {
#	$self->writedebug("Executing external function to get polarisation offset");
#	$self->{'_pol_offset'} = $ext_pol->($self);
#	$ocv += $self->{'_pol_offset'};
#    }
    if (ref($ext_pol) eq 'RemoteExec')
    {
	$self->writedebug("Executing external function to get polarisation offset");
	$self->{'_pol_offset'} = $ext_pol->exec_id($self);
	$ocv += $self->{'_pol_offset'};
    }
    my $step = abs(2 * $pol);
    my @plist = ($pol,-$pol,$pol,-$pol);
    foreach my $pol (@plist)
    {
	$pol += $ocv;
	if ($self->{'_mode'} =~ /A/)
	{
	    $pol = $self->{'_minpA'} if ($pol < $self->{'_minpA'});
	    $pol = $self->{'_maxpA'} if ($pol > $self->{'_maxpA'});
	}
	else
	{
	    $pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
	    $pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'}); 
	}
    }
    my @swp = ('A','B','C','D');
    for (my $x = 0; $x < @plist; $x++)
    {
	if ($self->{'_mode'} eq 'A')
	{
	    $self->checkcmd('K'.$swp[$x].$plist[$x]);
	}
	else
	{
	    $self->checkcmd('S'.$swp[$x].$plist[$x]);
	}
    }
    if ($self->{'_mode'} eq 'A')
    {
	$self->checkcmd(sprintf("IS %.4g",$step)); ## Stepsize ##
    }
    else
    {
	$self->checkcmd(sprintf("VS %.4g",$step)); ## Stepsize ##
    }
    $self->checkcmd('TE 0.01'); ## 10 ms steptime (50 Hz) ##
    my $sweeps = 1500;  ### output for 1 minute ###
    $self->writecmd('SM '.$sweeps);  ## Maximum of steps ##
    my $totaltime = $sweeps * 4 * 0.01 * $self->{'_overhead'}; ## total runtime is 100 seconds.. 
    return ($plist[0],$totaltime);
}



sub setup_sweep()
{
    my $self = shift;
    my $setup = shift;
    my $potlist = shift;
    die "Potential list not an arrayref\n" unless ($potlist && ref($potlist) =~ /ARRAY/);
    my @plist = @{$potlist};
    die "Too many independent potential points atttempted for potential sweep!\n" if (@plist > $self->{'_max_sweep_points'});
    my $sr = $self->get_setup_value(0,$setup,'sweep_rate');
    $self->writeerror("Refusing to set up a potential sweep with less than 2 points!") unless (@plist > 1);
    $self->setup_common($setup);
    my $ocv = $self->get_last_ocv();
    $ocv = 0 unless ($ocv);
    my $ext_pol = $self->get_setup_value($self->get_default('polarisation_offset'),$setup,'polarisation_offset');
#    if (ref($ext_pol) eq 'CODE')
#    {
#	$self->writedebug("Executing external function to get polarisation offset");
#	$self->{'_pol_offset'} = $ext_pol->($self);
#	$ocv += $self->{'_pol_offset'};
#    }
    if (ref($ext_pol) eq 'RemoteExec')
    {
	$self->writedebug("Executing external function to get polarisation offset");
	$self->{'_pol_offset'} = $ext_pol->exec_id($self);
	$ocv += $self->{'_pol_offset'};
    }
    foreach my $pol (@plist)
    {
	$pol += $ocv;
	if ($self->{'_mode'} =~ /A/)
	{
	    $pol = $self->{'_minpA'} if ($pol < $self->{'_minpA'});
	    $pol = $self->{'_maxpA'} if ($pol > $self->{'_maxpA'});
	}
	else
	{
	    $pol = $self->{'_minpV'} if ($pol < $self->{'_minpV'});
	    $pol = $self->{'_maxpV'} if ($pol > $self->{'_maxpV'}); 
	}
    }
    $sr = $self->{'_mins'} if ($sr < $self->{'_mins'});
    $sr = $self->{'_maxs'} if ($sr > $self->{'_maxs'});
    my $sp = $plist[0];
    my $totaltime = 0;
    my @tlist = ();
    for (my $x = 0; $x < scalar(@plist)-1; $x++)
    {
	my $t = abs($plist[$x+1] - $plist[$x]) / $sr;
	$t = 0.01 if ($t < 0.01);
	push @tlist,$t;
	$totaltime += $t;
    }
    if (@plist > 1)
    {
	push @tlist,0.01;
	$totaltime += 0.01;
    }
    my $sweeps = int($self->get_setup_value($self->get_default('sweeps'),$setup,'sweeps'));		     
    my $sweep_points = scalar(@plist)-1;
    die "Can not run potential sweep with more than one sweep with only 3 or less potential/current points!\n" if ($sweeps > 1 && @plist < 4);
    $sweep_points = 4 * $sweeps if ($sweeps > 1);
    $self->checkcmd('SM'.$sweep_points);
    my @swp = ('A','B','C','D');
    for (my $x = 0; $x < @plist; $x++)
    {
	if ($self->{'_mode'} eq 'A')
	{
	    $self->checkcmd('J'.$swp[$x].$plist[$x]);
	}
	else
	{
	    $self->checkcmd('V'.$swp[$x].$plist[$x]);
	}
	$self->checkcmd('T'.$swp[$x].$tlist[$x]);
    }
    my $keep_pot = $self->get_setup_value($self->get_default('keep_potential'),$setup,'keep_potential');
    my $kpot = 0;
    $kpot = 1 if ($keep_pot && $keep_pot =~ /yes/i);
    $self->checkcmd('OF'.$kpot);
    $self->checkcmd(sprintf("DL%.4g",$self->{'_delay'}));
    $totaltime = $totaltime * $self->{'_overhead'} * $sweeps;
    return ($plist[0],$totaltime);
}

sub start_sweep()
{
    my $self = shift;
    $self->writecmd('?PW');
    my $res = $self->readcmd();
    if (!$res || $res =~ /00/)
    {
	$self->checkcmd('PW1')
    }
    $self->writecmd('SW1'); ## Start sweep ##
    my $isok = 0;
    ### Wait for potentiostat to start sweep ###
    while (!$isok)
    {
	select(undef,undef,undef,0.2);
	$self->writecmd('?ST');
	my $res = $self->readcmd();
	if ($res && $res =~ /\d+/)
	{
	    $res += 0;
	    $isok = 1 if ($res > 2);
	}
    }
}

sub start_square()
{
    my $self = shift;
    $self->writecmd('?PW');
    my $res = $self->readcmd();
    if (!$res || $res =~ /00/)
    {
	$self->checkcmd('PW1')
    }
    $self->writecmd('SW2'); ## Start sweep ##
    my $isok = 0;
    ### Wait for potentiostat to start sweep ###
    while (!$isok)
    {
	select(undef,undef,undef,0.2);
	$self->writecmd('?ST');
	my $res = $self->readcmd();
	if ($res && $res =~ /\d+/)
	{
	    $res += 0;
	    $isok = 1 if ($res > 2);
	}
    }
}

sub setup_chrono()
{
    my $self = shift;
    my $setup = shift;
    my $truepol = $self->setup_common($setup);
    return $truepol;
}

sub start_chrono()
{
    my $self = shift;
    $self->writecmd('?PW');
    my $res = $self->readcmd();
    if (!$res || $res =~ /00/)
    {
	$self->checkcmd('PW1')
    }
}

sub setup_common()
{
    my $self = shift;
    my $setup = shift;
    $self->writecmd('OT3');
    $self->writecmd('GP2');
    $self->writecmd('OT1');  ### Terminator set to 'cr lf + EOI' ###
    $self->checkcmd('PW0') unless ($self->ignore());   ### off unless chrono is already running  ###
    $self->checkcmd('EB1'); ## beep off ##
    ### Measure range on input ###
    my $ra = $self->get_setup_value($self->get_default('range'),$setup,'range');
    $ra = 0 if ($ra =~ /auto/i);
    if ($ra)
    {
	my @list = @{$self->{'_ranges'}};
	my $tr = 0;  ## Default auto resistor ##
	for (my $x = 0; $x < @list; $x++)
	{
	    $tr = $x if ($list[$x] == $ra)
	}
	$ra = $tr;
    }
    $self->checkcmd('RG '.$ra);
    $self->checkcmd("FI0");  ### Low pass filter off ###
    $self->checkcmd('OT1');  ### Terminator set to 'cr lf + EOI' ###
    my $rr = $self->get_setup_value($self->get_default('measure_resistor'),$setup,'measure_resistor');
    $self->set_resistor($rr);
    $self->checkcmd('BY1');   ## Standby set to all o/c ##
    $self->checkcmd('ON0');   ## On mode set to pol V/I ##
    my $bandwidth = $self->get_setup_value($self->get_default('bandwidth'),$setup,'bandwidth');
    if ($bandwidth =~ /none/)
    {
	$self->checkcmd('SY1');   ## Bandwidth off ##
    }
    else
    {
	$self->checkcmd('SY0');   ## Bandwidth off ##
    }
    if ($bandwidth !~ /\d/)
    {
	my $tb = 2;
	my @l = $self->get_tag_values('bandwidth');
	for (my $x = 0; $x < @l; $x++)
	{
	    $tb = $x if ($bandwidth eq $l[$x]);
	}
	$bandwidth = $tb;
    }
    ### Default type C ###
    if ($self->{'_mode'} eq 'V')
    {
	die "Ilegal bandwidth '$bandwidth' selected for galvanostatic mode for $self->{'_type'}\n" unless ($bandwidth >= 0 && $bandwidth < 10); 
	$self->checkcmd('PO0');
	$self->checkcmd('PB'.$bandwidth);
    }
    else
    {
	$self->checkcmd('PO1');
	die "Ilegal bandwidth '$bandwidth' selected for galvanostatic mode for $self->{'_type'}\n" unless ($bandwidth >= 0 && $bandwidth < 3); 
	$self->checkcmd('GB'.$bandwidth);
    }
    my $pol = $self->get_setup_value(0,$setup,'polarisation');
    my $truepol = $self->set_polarisation($pol,$setup);
    $self->checkcmd('OL2');  ### warning if error ###
    $self->checkcmd('GP2');  ### sets gpib to short on ###
    $self->checkcmd('TR0');  ### sets trigger to single ###
    $self->checkcmd("BY0");  ### sets 1287 to standby ###
    $self->checkcmd("SW0");  ### Sweep off ###
    ### Current limit ###### 
    my $il = $self->get_setup_value($self->get_default('current_limit'),$setup,'current_limit');
    if ($il)
    { 
	my @list = @{$self->{'_values_current_limit'}};
	my $tl = 6; ### Defauylt current limit 
	for (my $x = 0; $x < @list; $x++)
	{
	    $tl = $x if ($list[$x] == $il);
	}
	$il = $tl;
    }
    else
    {
	$il = 6; ### Defauylt current limit 
    }
    $self->checkcmd('IL'.$il);
    $self->checkcmd('PX3'); ### set x to DeltaRE   ###
    $self->checkcmd('PY5'); ### set y to current   ###
    $self->checkcmd('UL3'); ### set displayX to DeltaRE ###
    $self->checkcmd('UR5'); ### set displayY to current ###
    $self->checkcmd('FD0'); ### set display ###
    my $pi = 0;
    $pi = 1 if ($self->get_setup_value($self->get_default('use_0.01_factor'),$setup,'use_0.01_factor') =~ /YES/i);
    $self->checkcmd('PI'.$pi);
    $self->{'_AC_gain'} = 1;
    $self->{'_AC_gain'} = 0.01 if ($pi);
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    return $truepol;
}

sub setup_impedance()
{
    my $self = shift;
    my $setup = shift;
    my $truepol = $self->setup_common($setup);
    $self->writecmd('?PW');
    my $res = $self->readcmd();
    if (!$res || $res =~ /00/)
    {
	$self->checkcmd('PW1')
    }
    $self->run_dvm();
    sleep(1);
    $self->run_dvm();
    sleep(1);
    $self->writecmd('RU1');  ### Run ###
    $self->checkcmd('BR0');
    my $br = $self->get_setup_value($self->get_default('bias_reject'),$setup,'bias_reject');
    if ($br && $br =~ /YES/i)
    {
	$self->checkcmd('VT0');
	$self->checkcmd('IT0');
	$self->checkcmd('BR1');
    }
    return $truepol;
}

sub run_dvm()
{
    my $self = shift;
    $self->writecmd('RU1');
    my $res = $self->readcmd();
    if ($res)
    {
	my @l = split(/\,/,$res);
	my $string = sprintf("%g %g",$l[0],$l[1]);
	push @{$self->{'_data'}},$string;
	return $string;
    }
    push @{$self->{'_data'}},'';
    return '';
}


sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### runs command twice if first go results in error ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    ### For some reason doesn't work although manual say it should... ###
    $self->writecmd('?ER');
    my $e = $self->readcmd();
    if (!$e || $e !~ /^00/)
    {
        if ($e = /~(\d+)/)
        {
            $self->writedebug('Error description: '.$self->errorcode($1));
        }
        $self->writecmd('CE');
        $self->writecmd($cmd);
        $self->checkerror();
    }
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('?ER');
    my $rep = $self->readcmd();
    if (!$rep || $rep !~ /^00/)
    {
        $self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '$rep'");
        if ($rep = /~(\d+)/)
        {
            $self->errorlog('Error description: '.$self->errorcode($1));
        }
        $self->writecmd('CE');
    }
}

sub get_error()
{
    my $self = shift;
    $self->writecmd('?ER');
    my $rep = $self->readcmd();
    if ($rep =~ /(\d+)/)
    {
	my $err = $1;
        my $r = $self->errorcode($1);
        $self->writedebug("Error description: $r");
        return $err;
    }
    return '';
}

package Solartron1286;

our @ISA = qw(Solartron1287);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Solartron 1286';
    $self->{'_minpV'} = -12.8;
    $self->{'_maxpV'} = 12.8;
    $self->{'_ranges'} = [0,0.1,2,20,200];
    $self->{'_values_range'} = ['auto',0.1,1,2,20,200];
    $self->{'_values_measure_resistor'} = ['auto',0.1,1,10,100,1000,10000,100000];
    $self->{'_resistors'} = [0,0.1,1,10,100,1000,10000,100000];
    bless $self,$class;
    return $self;
};







