package Stanford;
#$Id: Stanford.pm,v 1.19 2013/09/05 09:06:01 sofc Exp $
############################################################
#
#    Package Stanford
#
# This module implements the FRA class for Stanford Research Systems
# --{\textregistered} 
# SR830 Lock-In amplifiers.
#
# To get a device instance call one of the constructors:
#
# $fra = Stanford->new($gpib_address);
#
############################################################

use strict;
use warnings;
use FDEV;
use Math::Complex;
our @ISA = qw(FRA);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Stanford SR830';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V'};
    $self->{'_minf'} = 0.001;
    $self->{'_maxf'} = 102000;
    $self->{'_minpV'} = 0;
    $self->{'_maxpV'} = 0;
    $self->{'_minpA'} = undef;
    $self->{'_maxpA'} = undef;
    $self->{'_maxampA'} = undef;
    $self->{'_maxampV'} = 5;
    $self->{'_minampV'} = 0.004;
    $self->{'_source_modes'} = {'U/A' => 0, 'U/(A - B)' => 1, 'U/I (1 MOhm)' => 2,'U/I (100 MOhm)' => 3};
    $self->{'_status_errors'} = {
	'' => ' ',
	0 => ' ',
	1 => 'Backup error',
	2 => 'RAM error',
	3 => '',
	4 => 'ROM error',
	5 => 'GPIB error',
	6 => 'DSP error',
	7 => 'Math error',
    };
    $self->{'_errorcodes'} = {
	'' => ' ',
	0 => 'Input or amplifier overload', 
	1 => 'Time constant filter overload',
	2 => 'Output overload',
	3 => 'Reference unlock detected',
	4 => 'Frequency range change',
	5 => 'Time constant changed indirectly',
	6 => 'Data storage triggered',
	7 => ' ',
	'XX' => 'No data returned',
    };
    $self->{'_overhead'} = 1.05;
    $self->{'_minintegration'} = 1;
    $self->{'_setuptime'} = 6;
    $self->{'_delay'} = 0.1;  ## Default delay of 100 ms ##
    $self->{'_setup_tags'} = 'mode,overhead,amplitude,delay,harmonic,reference_trigger,outer_mode,input_coupling,notch_filter,sensitivity,reserve,time_constant,low_pass_filter,synchronous_filter,source_mode';
    $self->{'_values_source_mode'} = ['U/A','U/(A - B)','U/I (1 MOhm)','U/I (100 MOhm)'];
    $self->{'_values_reference_trigger'} = ['Zero crossing','TTL rising edge','TTL falling edge'];
    $self->{'_values_notch_filter'} = ['No filter','Line notch in','2x line notch in','Both notch fileters in'];
    $self->{'_values_harmonic'} = [1,2,3,4,5,6,7,8,9,10];
    $self->{'_values_outer_mode'} = ['Floating','Grounded'];
    $self->{'_values_input_coupling'} = ['AC','DC'];
    $self->{'_values_low_pass_filter'} = ['6 dB/oct','12 dB/okt','18 dB/okt','24 dB/okt'];
    $self->{'_values_reserve'} = ['High reserve','Normal','Low Noise'];
    $self->{'_values_synchronous_filter'} = ['Off','On below 200 Hz'];
    $self->{'_values_sensitivity'} = ['2 nV/fA','5 nV/fA','10 nV/fA','20 nV/fA','50 nV/fA','100 nV/fA','200 nV/fA','500 nV/fA','1 uV/pA','2 uV/pA','5 uV/pA','10 uV/pA','20 uV/pA','50 uV/pA','100 uV/pA','200 uV/pA','500 uV/pA','1 mV/nA','2 mV/nA','5 mV/nA','10 mV/nA','20 mV/nA','50 mV/nA','100 mV/nA','200 mV/nA','500 mV/nA','1 V/uA'];
    $self->{'_values_time_constant'} = ['10 us','30 us','100 us','300 us','1 ms','3 ms','10 ms','30 ms','100 ms','300 ms','1 s','3 s','10 s','30 s','100 s','300 s','1 ks','3 ks','10 ks','30 ks'];
    $self->{'_freq_seg_tags'} = ['sensitivity'];
    $self->{'_sensitivity'} = undef;
    $self->{'_default'} = {
        'mode' => 'Volt',
        'source_mode' => 'U/I (1 MOhm)',
        'polarisation_offset' => 'OFF',
        'reference_potential' => 'Fixed',
        'delay' => $self->{'_delay'},
        'amplitude' => $self->{'_minampV'},
        'polarisation' => 0,
	'reference_trigger' => 0,
	'outer_mode' => 'Grounded',
        'input_coupling' => 'DC',
	'harmonic' => 1,
	'notch_filter' => 'No filter',
	'sensitivity' => '1 V/uA',
	'reserve' => 'High reserve',
	'time_constant' => '1 ms',
	'low_pass_filter' => '6 dB/okt',
	'synchronous_filter' => 'On below 200 Hz',
	'ilin' => 0,
	'rmod' => 0,
	'sync' => 0,
	'ddef1' => '0,0',
	'ddef2' => '0,0',
	'fpop1' => 1,
	'fpop2' => 1,
	'oexp1' => '0,0',
	'oexp2' => '0,0',
    };
    $self->{'_desc'} = {
        'mode' => 'Control mode',
        'input_mode' => 'Input mode for the voltage input, either differential or single ended',
        'input_coupling' => 'Determines if the input coupling is direct through or uses a high pass filter',
        'harmonic' => 'Determines if the base frequency or one of the higher harmonics are used for measuring impedance',
        'polarisation_offset_program' => 'Select if a polarisation offset program (and which one) is to be run prior to impedance scans',
        'reference_potential' => 'Select if the polarisation is to be set versus chassis ground or versus measured potential at OCV',
        'delay' => 'Delay in seconds between each frequency',
        'amplitude' => 'AC amplitude (RMS, V)',
        'polarisation' => 'DC polarisation (V)',
        'source_mode' => 'Determines how the impedance is calculated',
        'outer_mode' => 'Determines if the screen on voltage input is grounded or left floating',
	'reference_trigger' => 'Sets trigger method',
	'notch_filter' => 'Selectsw which notch filter to use on the input',
	'sensitivity' => 'Sets device sensitivity (input range)',
	'reserve' => 'Sets filtering method',
	'time_constant' => 'Sets time constant on low pass filter',
	'low_pass_filter' => 'Sets low pass filter type',
	'synchronous_filter' => 'Select if a synchronous filter is on below 200 Hz',
    };
    $self->{'_save_tags'} = ['phas','fmod','freq','rslp','harm','slvl','isrc','ignd','icpl','ilin','sens','rmod','oflt','sync','ddef1','ddef2','fpop1','fpop2','oexp1','oexp2'];
    bless $self,$class;
    return $self;
}

return 1;

sub init()
{
    #### Do not use '*RST' command as it sets 1 V at 1kHz as default! ###
    my $self = shift;
    my $res = $self->reset;
    $self->checkcmd('PHAS0');
    $self->checkcmd('FMOD1');
    $self->checkcmd('FREQ10000');
    $self->checkcmd('RSLP0');
    $self->checkcmd('HARM1');
    $self->checkcmd('ISRC2');
    $self->checkcmd('IGND1');
    $self->checkcmd('ICPL1');
    $self->checkcmd('ILIN0');
    $self->checkcmd('SENS26');
    $self->checkcmd('RMOD0');
    $self->checkcmd('OFLT4');
    $self->checkcmd('OFSL0');
    $self->checkcmd('SYNC1');
    $self->checkcmd('REST');
    return $res;
}

sub reset()
{
    my $self = shift;
#    $self->writecmd('OUTX1');
    $self->checkcmd('SLVL'.$self->{'_minampV'});
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    $self->checkcmd('REST');
    return $res;
}

sub isallive()
{
    my $self = shift;
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /Stanford/i);
    return 0;
}

sub bulk_load_check(@)
{
    my $self = shift;
#    $self->writecmd('OUTX1');
    while (@_)
    {
        my $cmd = shift;
        $self->checkcmd($cmd);
    }
    return undef;
}

sub tag_value
{
    my $self = shift;
    my $key = shift;
    if ($key eq 'sensitivity')
    {
	return $self->{'_sensitivity'} unless (@_);
	my $cmd = 'SENS26'; ## Default 1 V/uA
	my $val = shift;
	my @list = @{$self->{'_values_sensitivity'}};
	for (my $x = 0; $x < @list; $x++)
	{
	    next unless ($val eq $list[$x]);
	    $cmd = 'SENS'.$x; 
	    $self->{'_sensitivity'} = $list[$x];
	    last;
	}
	$self->checkcmd($cmd);
	return 	$self->{'_sensitivity'};
    }
}

sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### runs command twice if first go results in error ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    $self->writecmd('ERRS?');
    my $e = $self->readcmd();
    if ($e)
    {
        if ($e = /~(\d+)/)
        {
	    my $r = $1;
            $self->writedebug('Error description: '.$self->status_code($r));
        }
        $self->writecmd('*CLS');
        $self->writecmd($cmd);
        $self->checkerror();
    }
}

sub get_error()
{
    my $self = shift;
    $self->writecmd('LIAS?');
    my $rep = $self->readcmd();
    if ($rep && $rep =~ /(\d+)/)
    {
	my $e = $1;
	my $r = $self->errorcode($e);
	$self->writedebug("Error description: $r");	    
	if (@_)
	{
	    my @l = ();
	    for my $x (0..7)
	    {
		$l[$x] = 0;
		$l[$x] = 1 if ($e & 2**$x);
	    }
	    # Specifically set $l[4] to 0 to ignore frequency range change 'errors'#
	    $l[4] = 0;
	    $e = 0;
	    for my $x (0..7)
	    {
		$e += 2**$x if ($l[$x]);
	    }
	    return $e;
	}
        return $rep;
    }
    return '';
}

sub status_code($)
{
    my $self = shift;
    my $val = shift;
    return '' unless ($val);
    $val += 0;
    my @l = ();
    for my $x (0..7)
    {
	$l[$x] = 0;
	$l[$x] = 1 if ($val & 2**$x);
    }
    my @r = ();
    for (0..7)
    {
	push @r, $self->{'_status_codes'}{$_} if ($l[$_]);
    }
    my $res = join(', ',@r);
    return $res
}


sub errorcode($)
{
    my $self = shift;
    my $val = shift;
    return '' unless ($val);
    if ($val !~ /^\d+$/)
    {
	return $self->{'_errorcodes'}{$val} if (exists($self->{'_errorcodes'}{$val}));
	return $val;
    }
    $val += 0;
    my @l = ();
    for my $x (0..7)
    {
	$l[$x] = 0;
	$l[$x] = 1 if ($val & 2**$x);
    }
    my @r = ();
    for my $x (0..7)
    {
	next if (@_ && $x == 4); ### On purpose ignoring 4 (Frequency range change) ###
	push @r, $self->{'_errorcodes'}{$x} if ($l[$x]);
    }
    my $res = join(', ',@r);
    return $res
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('ERRS?');
    my $rep = $self->readcmd();
    if ($rep)
    {
        $self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '".$self->status_code($rep)."'");
        $self->writecmd('*CLS');
    }
}

sub readvalue()
{
    my $self = shift;
    my $cmd = shift;
    my $tag = $cmd;
    if ($cmd =~  /^(\w+?)(\d+)/)
    {
	$cmd = $1.'?'.$2;
    }
    else
    {
	$cmd .= '?';
    }
    my $res = undef;
    my $count = 0;
    while (!defined($res) && $count++ < 5)
    {
        $self->writecmd($cmd);
        my $r = $self->readcmd();
        if ($r =~ /([\d\,\.\+\-eE]+)/)
        {
            $res = $1;
        }
    }
    return $res if (defined($res));
    $count--;
    my $str = "Could not read value for tag '$tag', tried $count times!";
    $self->errorlog($str);
    return undef;
}

sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
        push @res,$tag.$self->readvalue($tag);
    }
    return @res;
}

sub setup_impedance($)
{
    my $self = shift;
    my $setup = shift;
    $self->init();
    $self->checkcmd('PHAS0');     ### Set phase to 0;
    $self->checkcmd('FMOD1');     ### Internal reference
    $self->checkcmd('OFLT4');     ### Time constant 10 us, to be overridden by individual frequency measurement ###
    $self->checkcmd('DDEF1,0,0');
    $self->checkcmd('DDEF2,0,0');
    $self->checkcmd('FPOP1,1');
    $self->checkcmd('FPOP2,1');
    $self->checkcmd('OEXP1,0,0');
    $self->checkcmd('OEXP2,0,0');
    my $sm = $self->get_setup_value($self->get_default('source_mode'),$setup,'source_mode');
    $self->{'_source_mode'} = $sm;
    $self->checkcmd('ISRC'.$self->{'_source_modes'}{$sm});
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    $self->{'_delay'} = 0 unless (defined($self->{'_delay'}) && $self->{'_delay'} > 0);
    my $harmonic = $self->get_setup_value($self->get_default('harmonic'),$setup,'harmonic');
    $self->checkcmd('HARM'.$harmonic);
    my $ou = 1;  ## Default Ground
    foreach my $x (0..1)
    {
	$ou = $x if ($self->{'_values_outer_mode'}[$x] eq $self->get_setup_value($self->get_default('outer_mode'),$setup,'outer_mode'));
    }   
    $self->checkcmd('IGND'.$ou);
    my $ac = 1; ## Default DC
    foreach my $x (0..1)
    {
	$ac = $x if ($self->{'_values_input_coupling'}[$x] eq $self->get_setup_value($self->get_default('input_coupling'),$setup,'input_coupling'));	
    }   
    $self->checkcmd('ICPL'.$ac);
    my $notch = 0;
    foreach my $x (0..3)
    {
	$notch = $x if ($self->{'_values_notch_filter'}[$x] eq $self->get_setup_value($self->get_default('notch_filter'),$setup,'notch_filter'));	
    }   
    $self->checkcmd('ILIN'.$notch);
    my $fi = 0; 
    foreach my $x (0..3)
    {
	$fi = $x if ($self->{'_values_low_pass_filter'}[$x] eq $self->get_setup_value($self->get_default('low_pass_filter'),$setup,'low_pass_filter'));	
    }   
    $self->checkcmd('OFSL'.$fi);
    my $reserv = 0;
    foreach my $x (0..2)
    {
	$reserv = $x if ($self->{'_values_reserve'}[$x] eq $self->get_setup_value($self->get_default('reserve'),$setup,'reserve'));	
    }   
    $self->checkcmd('RMOD'.$reserv);
    my $trig = 0;
    foreach my $x (0..2)
    {
	$trig = $x if ($self->{'_values_reference_trigger'}[$x] eq $self->get_setup_value($self->get_default('reference_trigger'),$setup,'reference_trigger'));	
    }   
    $self->checkcmd('RSLP'.$reserv);
    my $sync = 0;
    foreach my $x (0..1)
    {
	$sync = $x if ($self->{'_values_synchronous_filter'}[$x] eq $self->get_setup_value($self->get_default('synchronous_filter'),$setup,'synchronous_filter'));	
    }   
    $self->checkcmd('SYNC'.$sync);
    my $sens = 26; ### Default sens of 1V
    foreach my $x (0..26)
    {
	$sens = $x if ($self->{'_values_sensitivity'}[$x] eq $self->get_setup_value($self->get_default('sensitivity'),$setup,'sensitivity'));	
    }
    $self->checkcmd('SENS'.$sens);  ### May be overridden by individual frequency measurement ###
    $self->{'_sensitivity'} = $self->get_setup_value($self->get_default('sensitivity'),$setup,'sensitivity');
    $self->amplitude($self->get_setup_value($self->{'_minampV'},$setup,'amplitude'));
    $self->reset_data;
    $self->get_error('ignore');
    $self->checkcmd('REST');
    return 0; ### No DC polarisation possible
}


sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $amp = shift;
    $amp = $self->{'_minampV'} if ($amp < $self->{'_minampV'});
    $amp = $self->{'_maxampV'} if ($amp > $self->{'_maxampV'});
    $self->checkcmd(sprintf("SLVL%.4g",$amp));  ## Voltage AC amplitude ##
    $self->{'_amplitude'} = $amp;
    return $self->{'_amplitude'};
}


sub run_frequency($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    $time = $self->get_integration_time($freq,$time);
    my $fout = sprintf("%.4g",$freq);
    $self->checkcmd('FREQ'.$fout);
    # Select correct filter ###
    my $t = 0;
    $t = 1/$freq if ($freq);
    $t = $time if ($t < $time);
    $t = $t / 5; ## Sampling occours over the last 4-5 time constants ##
    my @sens = @{$self->{'_values_time_constant'}};
    my @sreal = ();
    for (my $x = 0; $x < @sens; $x++)
    {
	$sens[$x] =~ /(\d+)\s+(\w+)/;
	my $val = $1;
	my $exp = $2;
	$val = $val * 1e-6 if ($exp eq 'us');
	$val = $val * 1e-3 if ($exp eq 'ms');
	$val = $val * 1e3 if ($exp eq 'ks');
	$sreal[$x] = $val;
    }
    my $tc = 0;
    for (my $x = 0; $x < @sreal; $x++)
    {
	$tc = $x if ($sreal[$x] < $t);
    }
    $self->checkcmd('OFLT'.$tc);
    $self->checkcmd('REST');
    sleep($self->{'_delay'});
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->();
        }
        next if (!$testfile || -e $testfile);
        $self->reset();
        push @{$self->{'_data'}},undef;
        return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
        my $readloop = 0;
      INNER:
        while ($readloop < 5)
        {
            $self->writecmd('SNAP?1,2');
            my $res = $self->readcmd();
            if (!$res)
            {
                $readloop++;
                next INNER;
            }
	    $self->checkcmd('REST');
            my @r = split(",",$res);
	    my $result = sprintf("%f %f %f",$freq,$r[0]*$f,$r[1]*$f);
	    if ($self->{'_source_mode'} =~ /I/)
	    {
		my $ampl = cplx($self->amplitude,0);
		my $i = cplx(-$r[0],-$r[1]);
		my $z = cplx(0,0);
		my $r = cplx(1000,0);
		my $du = $ampl - $i * $r;
		$self->{'_dU'} = $du;
		$self->{'_I'} = $i;
		if (abs($i))
		{
		    $z = $du/$i; 
		}
		$result = sprintf("%f %f %f",$freq,Re($z)*$f,Im($z)*$f);
	    }
            push @{$self->{'_data'}},$result;
	    my $error = $self->get_error('ignore');
            $self->{'_last_error'} = $error if ($error);
            return $result;
        }
        sleep(1);
        if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next OUTER if (!$testfile || -e $testfile);
        $self->reset();
        push @{$self->{'_data'}},undef;
        $self->{'_last_error'} = 'XX';
        return undef;
    }
    push @{$self->{'_data'}},undef;
    $self->{'_last_error'} = 'XX';
    return undef;
}


sub run_frequency_all($$)
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    if ($self->{'_source_mode'} =~ /I/)
    {
	$result .= ' # deltaU: '.$self->{'_dU'}.', I: '.$self->{'_I'};
	my $sn = Re($self->{'_dU'})*100/$self->amplitude;
	$result .= sprintf(", %.2f%% of output applied to sample",$sn);
    }
    push @{$self->{'_longdata'}},$result;	
    return $result;
}
